Mac instructions:
	Go buy a windows machine.

Windows instructions:

Problem / Assumptions / Scope

	The purpose of this tool is to enable non technical content producers, piggybacking on the familiarity and convenience of google spreadsheets for database content.
	Given a google folder id (or single spreadsheet id), it will recreate an entire google directory structure in your local folder system.
	Row 1 in your spreadsheet will become your json keys - it is recommended to lock this row
	You can use frozen rows or columns without a value for notes and instructions, but in general the spreadsheet should look like a table:
	 ________________________
	|___|__A__|__B___|___C___|
	|_1_|_id__|_name_|_color_| 
	|_2_|__1__|_Fred_|__red__|
	|_3_|__2__|_Jake_|_blue__|

	would output a json list:
	[
        {
            "id": "1",
            "name": "Fred",
            "color": "red"
        },
        {
            "id": "2",
            "name": "Jake",
            "color": "blue"
        }
	]

	In order to interface with Google, you will need to enable the Google Drive API, register a project in your drive, and complete OAuth authenitication.
	Some familiarity with command line, python, paths, json, google, and the internet is expected; but nothing beyond an entry level computer science class.  

Quick Start

	Install Python 3.4 or higher (tool was developed against Python 3.7.1) https://www.python.org/downloads/
		(You might also need to make python.exe part of your system's path environment variable, Python 3.7 and later have install options to do this for you)

	Install Pip (should be packaged with python 3 and above)
	
	Install Google API:
		pip install --upgrade google-api-python-client --user
		pip install --upgrade google-api-python-client oauth2client --user

	Enable the google sheets API
		https://developers.google.com/sheets/api/quickstart/python
		"Click this button to create a new Cloud Platform project and automatically enable the Google Sheets API"
		You might get a 'create project dialog' dialog if this is your first time doing this on your google account
		A dialog should prompt with "Download Client Configuration"
		You will get Client ID and Client Secret that looks something like this
		Client ID 458213654518-l35tni155bov9oovoa8hms1u277c4nhi.apps.googleusercontent.com 
		Client Secret 9niT3kshg_70kUsjS1n7C6gS
		Download it as "credentials.json", save it local to the tool.

	Run python driveFileLister.py from this directory to get a token and verify that you are properly configured
		A new browser tab should open, you might get a "do you trust this app" security warning, select advanced and proceed 
		Next, you will be ask for specific permissions - you are safe, we are just asking to read data,
		finally there will be a file printout, and make a token.json file made in the project directory so that your computer is now authorized
		As a developer, you may find it necessary to delete the token.json file if you change google file access privileges (read / write access)
		This token doesn't want to generate if scripts are ran from outside of the local directory

	If you have never enabled Google APIs before, you will likely get a 403 error directing you to the API console
		This is for a global setting of Google Drive API, while the rest of our steps have just been authenticating our "app"
		https://console.developers.google.com/apis/api/drive.googleapis.com/overview?project=458213654518 or similar will be in the error
		You will need to navigate there and enable the API.

	Customize the config file for your project, most importantly the id of the directory or spreadsheet needs to be set
		From a url like https://docs.google.com/spreadsheets/d/1wlUGddksS2iTq3IT_ChZtWsun5bC8xh_NVNS_hzajag/edit#gid=0
		or https://drive.google.com/drive/u/0/folders/1o6ei-cJ3WSz0xfAi6MGHC2ew2M6cyHPi
		the long series of characters is the google id "1wlUGddksS2iTq3IT_ChZtWsun5bC8xh_NVNS_hzajag" and "1o6ei-cJ3WSz0xfAi6MGHC2ew2M6cyHPi"
	
	Run the tool: python googleContentPull.py (defaults to config file found locally)

	In the ideal setup, this tool is brought into a repo as a submodule. 
		Copy the config file to your project directory and check it into source control, then pass the config path as a parameter
		(You might find it convenient to write a batch file for this command, this is an example of the config file and submodule living in a parent directory)
		python ./googlesheettojson/googleContentPull.py ./config.json


Instructions for the Config file

	"googleParentFolderID": "",           	The id of the google drive folder, it will search the entire google directory for spreadsheets
	"localDirectoryBasePath": "./",			The local path where you want json files to be made, a folder will be made for every google folder, and 
	"numResults": 1000,						Google limits how many files are returned in queries, you need to make this number sufficiently large to capture all your files

	"allowGapRows": true,					This will break link backs! - How to handle an empty row in the spreadsheet, True skips the row and continues, False terminates generating json when hitting an empty row
	"allowGapColumns": true,				How to handle an empty column, this only affects the first row, True skips an empty value for the first row, False terminates making the schema if empty cell is encountered
	"enforceASCII": false,					True will escape utf8 characters, False will keep values as they appear in the sheet

	"jsonExtension": ".json",				There might be a rare case where you need a custom extension
	"pathSlash": "/",						Supporting different file systems
	"outputEncoding": "utf8",				I would think this would always be utf8, but someone might need ascii or another encoding

	"logLevel": 0,							The tool logs things, 0 = everything, 1 = warnings & errors, 2 = errors only
	"logOutputFile": "somePath.txt" 		The name you want to use for your log file, you can specify an entire path starting from the tool directory


Limitations

	Prerequisites

		Python 3.2 or higher for operating system calls

	Google drive file rules are different that OS file rules

	* Google files can have the same name in the same folder since google files operate based on hidden id#
	* File names should never contain spaces
	* There are different file name character rules per operating system, avoid all of the following:
		* / \ ? % : | " > < . *
	* There is a max path length in Windows, avoid deep folder nesting and long names

	Google Drive OS throttles how many files are returned in queries, you might need to provide filtering, # of files, or a search query to limit results

	Google doesn't really have directories even though it displays folders, there is a lot of recursion in code to create and traverse folders, recursion obviously has a stack limit.

	Google Drive files must not be private - they must have some public share setting or link sharing

	This tool just pulls down files and converts to json, it does not do anything "smart" comparing what files are local to what is remote. For example, if you delete a file in google, running this tool will not delete the local file.

	While Google Drive updates immediately, it actually takes a few minutes for files to be deleted, or for files to be moved into folders, or a new folder to be recognized.

	All values in the sheet are exported as strings, this was a conscious decision, this tool's job is not type checking, validation, or sanitizing whitespace.
	Retrieving content is a generic problem, while validation is project specific. 

References

	Google Scopes URLs
	https://developers.google.com/identity/protocols/googlescopes

	Google Drive Search Parameters
	https://developers.google.com/drive/api/v3/search-parameters

	Google Drive MIME types
	https://developers.google.com/drive/api/v3/mime-types

	Python v3 API
	https://docs.python.org/3/library/
	
	Google Sheets API
	https://developers.google.com/resources/api-libraries/documentation/sheets/v4/python/latest/sheets_v4.spreadsheets.html

	Google Drive API
	https://developers.google.com/resources/api-libraries/documentation/drive/v3/python/latest/drive_v3.files.html

	Python writing to JSON
	https://stackabuse.com/reading-and-writing-json-to-a-file-in-python/

	If you have the file id, the sheet id, and know the column and row, you can link directly to that cell using the sheet id for gid and column, row for range value
	e.g.
	https://docs.google.com/spreadsheets/d/13PuCx8zKUjXvofFYBGzoOYog7UHpvLzCgxMLF9INnr8/edit#gid=0&range=D10


Instructions Google Drive Python API v3

	Prerequisites

		Python 2.7, or 3.4 or higher

	https://developers.google.com/api-client-library/python/apis/drive/v3

	pip install --upgrade google-api-python-client --user

	Setup Validation

	The Google Python file "driveFileLister.py" can be used to prove that you are correctly configured
		Open a command line and execute the following command:
		python driveFileLister.py
	It will take a moment to run, if you have more than 1000 spreadsheets it will not list them all

	You might get a warning that the API is not enabled, and will be directed to a link to enable it
	https://console.developers.google.com/apis/api/drive.googleapis.com/overview?project=458213654518


Instructions Google Sheet Python API (this just pulls down Sheet objects from Google)

	Prerequisites

	    Python 2.6 or greater --> Tool was developed with Python 3.7.1, not all versions of python have been tested 
	    The pip package management tool
	    A Google account

	Enable the google sheets API

		https://developers.google.com/sheets/api/quickstart/python

		You will get Client ID and Client Secret that looks something like this
		Client ID 458213654518-l35tni155bov9oovoa8hms1u277c4nhi.apps.googleusercontent.com 
		Client Secret 9niT3kshg_70kUsjS1n7C6gS

		download it as credentials.json, save it local to the tool.

	For the pip install, I have always needed to add the "--user" flag

		pip install --upgrade google-api-python-client oauth2client --user


Developer Reference: Sample Results

The metadata exported is useful in creating hyperlinks back to google, see references 

	{
    "metadata": {
        "spreadsheetUrl": "https://docs.google.com/spreadsheets/d/1wlUGddksS2iTq3IT_ChZtWsun5bC8xh_NVNS_hzajag/edit",
        "title": "Sheet1",
        "sheetId": 0,
        "columnCount": 26,
        "rowCount": 1001,
        "frozenRowCount": 1,
        "schema": [
            "id",
            "column_0",
            "column_1",
            "column_2"
        ],
        "schema_column": [
            0,
            1,
            2,
            4
        ]
    },
    "data": [
        {
            "id": "1",
            "column_0": "0",
            "column_1": "1",
            "column_2": "2"
        },
        {
            "id": "2",
            "column_0": "0",
            "column_2": "2"
        },
        {
            "id": "3",
            "column_0": "0",
            "column_1": "1",
            "column_2": "3"
        },
        {
            "column_0": "0",
            "column_1": "1",
            "column_2": "3"
        },
        {
            "id": "4",
            "column_0": "0",
            "column_1": "1",
            "column_2": "2"
        },
        {
            "id": "5",
            "column_0": "\"0\", \"column_1\": \"1\"",
            "column_1": "column_2': '2'"
        }
    ]
}

Developer Reference: Google Drive File Object

	Taken from: https://developers.google.com/resources/api-libraries/documentation/drive/v3/python/latest/drive_v3.files.html#list

	{ # A list of files.
    "files": [ # The list of files. If nextPageToken is populated, then this list may be incomplete and an additional page of results should be fetched.
      { # The metadata for a file.
        "hasThumbnail": True or False, # Whether this file has a thumbnail. This does not indicate whether the requesting app has access to the thumbnail. To check access, look for the presence of the thumbnailLink field.
        "mimeType": "A String", # The MIME type of the file.
            # Drive will attempt to automatically detect an appropriate value from uploaded content if no value is provided. The value cannot be changed unless a new revision is uploaded.
            # If a file is created with a Google Doc MIME type, the uploaded content will be imported if possible. The supported import formats are published in the About resource.
        "modifiedByMeTime": "A String", # The last time the file was modified by the user (RFC 3339 date-time).
        "thumbnailLink": "A String", # A short-lived link to the file's thumbnail, if available. Typically lasts on the order of hours. Only populated when the requesting app can access the file's content.
        "thumbnailVersion": "A String", # The thumbnail version for use in thumbnail cache invalidation.
        "explicitlyTrashed": True or False, # Whether the file has been explicitly trashed, as opposed to recursively trashed from a parent folder.
        "isAppAuthorized": True or False, # Whether the file was created or opened by the requesting app.
        "teamDriveId": "A String", # ID of the Team Drive the file resides in.
        "writersCanShare": True or False, # Whether users with only writer permission can modify the file's permissions. Not populated for Team Drive files.
        "ownedByMe": True or False, # Whether the user owns the file. Not populated for Team Drive files.
        "viewedByMeTime": "A String", # The last time the file was viewed by the user (RFC 3339 date-time).
        "id": "A String", # The ID of the file.
        "sharingUser": { # Information about a Drive user. # The user who shared the file with the requesting user, if applicable.
          "me": True or False, # Whether this user is the requesting user.
          "kind": "drive#user", # Identifies what kind of resource this is. Value: the fixed string "drive#user".
          "displayName": "A String", # A plain text displayable name for this user.
          "permissionId": "A String", # The user's ID as visible in Permission resources.
          "emailAddress": "A String", # The email address of the user. This may not be present in certain contexts if the user has not made their email address visible to the requester.
          "photoLink": "A String", # A link to the user's profile photo, if available.
        },
        "size": "A String", # The size of the file's content in bytes. This is only applicable to files with binary content in Drive.
        "videoMediaMetadata": { # Additional metadata about video media. This may not be available immediately upon upload.
          "width": 42, # The width of the video in pixels.
          "durationMillis": "A String", # The duration of the video in milliseconds.
          "height": 42, # The height of the video in pixels.
        },
        "lastModifyingUser": { # Information about a Drive user. # The last user to modify the file.
          "me": True or False, # Whether this user is the requesting user.
          "kind": "drive#user", # Identifies what kind of resource this is. Value: the fixed string "drive#user".
          "displayName": "A String", # A plain text displayable name for this user.
          "permissionId": "A String", # The user's ID as visible in Permission resources.
          "emailAddress": "A String", # The email address of the user. This may not be present in certain contexts if the user has not made their email address visible to the requester.
          "photoLink": "A String", # A link to the user's profile photo, if available.
        },
        "folderColorRgb": "A String", # The color for a folder as an RGB hex string. The supported colors are published in the folderColorPalette field of the About resource.
            # If an unsupported color is specified, the closest color in the palette will be used instead.
        "appProperties": { # A collection of arbitrary key-value pairs which are private to the requesting app.
            # Entries with null values are cleared in update and copy requests.
          "a_key": "A String",
        },
        "capabilities": { # Capabilities the current user has on this file. Each capability corresponds to a fine-grained action that a user may take.
          "canUntrash": True or False, # Whether the current user can restore this file from trash.
          "canMoveItemWithinTeamDrive": True or False, # Whether the current user can move this Team Drive item within this Team Drive. Note that a request to change the parent of the item may still fail depending on the new parent that is being added. Only populated for Team Drive items.
          "canDeleteChildren": True or False, # Whether the current user can delete children of this folder. This is false when the item is not a folder. Only populated for Team Drive items.
          "canMoveChildrenWithinTeamDrive": True or False, # Whether the current user can move children of this folder within the Team Drive. This is false when the item is not a folder. Only populated for Team Drive items.
          "canListChildren": True or False, # Whether the current user can list the children of this folder. This is always false when the item is not a folder.
          "canRename": True or False, # Whether the current user can rename this file.
          "canAddChildren": True or False, # Whether the current user can add children to this folder. This is always false when the item is not a folder.
          "canShare": True or False, # Whether the current user can modify the sharing settings for this file.
          "canTrashChildren": True or False, # Whether the current user can trash children of this folder. This is false when the item is not a folder. Only populated for Team Drive items.
          "canReadRevisions": True or False, # Whether the current user can read the revisions resource of this file. For a Team Drive item, whether revisions of non-folder descendants of this item, or this item itself if it is not a folder, can be read.
          "canCopy": True or False, # Whether the current user can copy this file. For a Team Drive item, whether the current user can copy non-folder descendants of this item, or this item itself if it is not a folder.
          "canMoveItemIntoTeamDrive": True or False, # Whether the current user can move this item into a Team Drive. If the item is in a Team Drive, this field is equivalent to canMoveTeamDriveItem.
          "canMoveItemOutOfTeamDrive": True or False, # Whether the current user can move this Team Drive item outside of this Team Drive by changing its parent. Note that a request to change the parent of the item may still fail depending on the new parent that is being added. Only populated for Team Drive items.
          "canComment": True or False, # Whether the current user can comment on this file.
          "canChangeViewersCanCopyContent": True or False, # Deprecated
          "canTrash": True or False, # Whether the current user can move this file to trash.
          "canDelete": True or False, # Whether the current user can delete this file.
          "canMoveTeamDriveItem": True or False, # Deprecated - use canMoveItemWithinTeamDrive or canMoveItemOutOfTeamDrive instead.
          "canDownload": True or False, # Whether the current user can download this file.
          "canChangeCopyRequiresWriterPermission": True or False, # Whether the current user can change the copyRequiresWriterPermission restriction of this file.
          "canMoveChildrenOutOfTeamDrive": True or False, # Whether the current user can move children of this folder outside of the Team Drive. This is false when the item is not a folder. Only populated for Team Drive items.
          "canRemoveChildren": True or False, # Whether the current user can remove children from this folder. This is always false when the item is not a folder. For Team Drive items, use canDeleteChildren or canTrashChildren instead.
          "canReadTeamDrive": True or False, # Whether the current user can read the Team Drive to which this file belongs. Only populated for Team Drive files.
          "canEdit": True or False, # Whether the current user can edit this file.
        },
        "trashedTime": "A String", # The time that the item was trashed (RFC 3339 date-time). Only populated for Team Drive files.
        "webViewLink": "A String", # A link for opening the file in a relevant Google editor or viewer in a browser.
        "version": "A String", # A monotonically increasing version number for the file. This reflects every change made to the file on the server, even those not visible to the user.
        "parents": [ # The IDs of the parent folders which contain the file.
            # If not specified as part of a create request, the file will be placed directly in the user's My Drive folder. If not specified as part of a copy request, the file will inherit any discoverable parents of the source file. Update requests must use the addParents and removeParents parameters to modify the parents list.
          "A String",
        ],
        "sharedWithMeTime": "A String", # The time at which the file was shared with the user, if applicable (RFC 3339 date-time).
        "shared": True or False, # Whether the file has been shared. Not populated for Team Drive files.
        "copyRequiresWriterPermission": True or False, # Whether the options to copy, print, or download this file, should be disabled for readers and commenters.
        "fullFileExtension": "A String", # The full file extension extracted from the name field. May contain multiple concatenated extensions, such as "tar.gz". This is only available for files with binary content in Drive.
            # This is automatically updated when the name field changes, however it is not cleared if the new name does not contain a valid extension.
        "originalFilename": "A String", # The original filename of the uploaded content if available, or else the original value of the name field. This is only available for files with binary content in Drive.
        "viewersCanCopyContent": True or False, # Deprecated - use copyRequiresWriterPermission instead.
        "description": "A String", # A short description of the file.
        "modifiedTime": "A String", # The last time the file was modified by anyone (RFC 3339 date-time).
            # Note that setting modifiedTime will also update modifiedByMeTime for the user.
        "viewedByMe": True or False, # Whether the file has been viewed by this user.
        "modifiedByMe": True or False, # Whether the file has been modified by this user.
        "owners": [ # The owners of the file. Currently, only certain legacy files may have more than one owner. Not populated for Team Drive files.
          { # Information about a Drive user.
            "me": True or False, # Whether this user is the requesting user.
            "kind": "drive#user", # Identifies what kind of resource this is. Value: the fixed string "drive#user".
            "displayName": "A String", # A plain text displayable name for this user.
            "permissionId": "A String", # The user's ID as visible in Permission resources.
            "emailAddress": "A String", # The email address of the user. This may not be present in certain contexts if the user has not made their email address visible to the requester.
            "photoLink": "A String", # A link to the user's profile photo, if available.
          },
        ],
        "createdTime": "A String", # The time at which the file was created (RFC 3339 date-time).
        "quotaBytesUsed": "A String", # The number of storage quota bytes used by the file. This includes the head revision as well as previous revisions with keepForever enabled.
        "starred": True or False, # Whether the user has starred the file.
        "properties": { # A collection of arbitrary key-value pairs which are visible to all apps.
            # Entries with null values are cleared in update and copy requests.
          "a_key": "A String",
        },
        "md5Checksum": "A String", # The MD5 checksum for the content of the file. This is only applicable to files with binary content in Drive.
        "iconLink": "A String", # A static, unauthenticated link to the file's icon.
        "imageMediaMetadata": { # Additional metadata about image media, if available.
          "exposureBias": 3.14, # The exposure bias of the photo (APEX value).
          "exposureTime": 3.14, # The length of the exposure, in seconds.
          "cameraMake": "A String", # The make of the camera used to create the photo.
          "maxApertureValue": 3.14, # The smallest f-number of the lens at the focal length used to create the photo (APEX value).
          "width": 42, # The width of the image in pixels.
          "focalLength": 3.14, # The focal length used to create the photo, in millimeters.
          "exposureMode": "A String", # The exposure mode used to create the photo.
          "colorSpace": "A String", # The color space of the photo.
          "location": { # Geographic location information stored in the image.
            "latitude": 3.14, # The latitude stored in the image.
            "altitude": 3.14, # The altitude stored in the image.
            "longitude": 3.14, # The longitude stored in the image.
          },
          "subjectDistance": 42, # The distance to the subject of the photo, in meters.
          "height": 42, # The height of the image in pixels.
          "lens": "A String", # The lens used to create the photo.
          "isoSpeed": 42, # The ISO speed used to create the photo.
          "meteringMode": "A String", # The metering mode used to create the photo.
          "flashUsed": True or False, # Whether a flash was used to create the photo.
          "time": "A String", # The date and time the photo was taken (EXIF DateTime).
          "aperture": 3.14, # The aperture used to create the photo (f-number).
          "rotation": 42, # The rotation in clockwise degrees from the image's original orientation.
          "sensor": "A String", # The type of sensor used to create the photo.
          "whiteBalance": "A String", # The white balance mode used to create the photo.
          "cameraModel": "A String", # The model of the camera used to create the photo.
        },
        "kind": "drive#file", # Identifies what kind of resource this is. Value: the fixed string "drive#file".
        "name": "A String", # The name of the file. This is not necessarily unique within a folder. Note that for immutable items such as the top level folders of Team Drives, My Drive root folder, and Application Data folder the name is constant.
        "webContentLink": "A String", # A link for downloading the content of the file in a browser. This is only available for files with binary content in Drive.
        "trashingUser": { # Information about a Drive user. # If the file has been explicitly trashed, the user who trashed it. Only populated for Team Drive files.
          "me": True or False, # Whether this user is the requesting user.
          "kind": "drive#user", # Identifies what kind of resource this is. Value: the fixed string "drive#user".
          "displayName": "A String", # A plain text displayable name for this user.
          "permissionId": "A String", # The user's ID as visible in Permission resources.
          "emailAddress": "A String", # The email address of the user. This may not be present in certain contexts if the user has not made their email address visible to the requester.
          "photoLink": "A String", # A link to the user's profile photo, if available.
        },
        "spaces": [ # The list of spaces which contain the file. The currently supported values are 'drive', 'appDataFolder' and 'photos'.
          "A String",
        ],
        "permissionIds": [ # List of permission IDs for users with access to this file.
          "A String",
        ],
        "trashed": True or False, # Whether the file has been trashed, either explicitly or from a trashed parent folder. Only the owner may trash a file, and other users cannot see files in the owner's trash.
        "contentHints": { # Additional information about the content of the file. These fields are never populated in responses.
          "indexableText": "A String", # Text to be indexed for the file to improve fullText queries. This is limited to 128KB in length and may contain HTML elements.
          "thumbnail": { # A thumbnail for the file. This will only be used if Drive cannot generate a standard thumbnail.
            "mimeType": "A String", # The MIME type of the thumbnail.
            "image": "A String", # The thumbnail data encoded with URL-safe Base64 (RFC 4648 section 5).
          },
        },
        "fileExtension": "A String", # The final component of fullFileExtension. This is only available for files with binary content in Drive.
        "hasAugmentedPermissions": True or False, # Whether any users are granted file access directly on this file. This field is only populated for Team Drive files.
        "permissions": [ # The full list of permissions for the file. This is only available if the requesting user can share the file. Not populated for Team Drive files.
          { # A permission for a file. A permission grants a user, group, domain or the world access to a file or a folder hierarchy.
            "domain": "A String", # The domain to which this permission refers.
            "displayName": "A String", # A displayable name for users, groups or domains.
            "teamDrivePermissionDetails": [ # Details of whether the permissions on this Team Drive item are inherited or directly on this item. This is an output-only field which is present only for Team Drive items.
              {
                "inheritedFrom": "A String", # The ID of the item from which this permission is inherited. This is an output-only field and is only populated for members of the Team Drive.
                "role": "A String", # The primary role for this user. While new values may be added in the future, the following are currently possible:
                    # - organizer
                    # - fileOrganizer
                    # - writer
                    # - commenter
                    # - reader
                "teamDrivePermissionType": "A String", # The Team Drive permission type for this user. While new values may be added in future, the following are currently possible:
                    # - file
                    # - member
                "inherited": True or False, # Whether this permission is inherited. This field is always populated. This is an output-only field.
              },
            ],
            "allowFileDiscovery": True or False, # Whether the permission allows the file to be discovered through search. This is only applicable for permissions of type domain or anyone.
            "deleted": True or False, # Whether the account associated with this permission has been deleted. This field only pertains to user and group permissions.
            "kind": "drive#permission", # Identifies what kind of resource this is. Value: the fixed string "drive#permission".
            "emailAddress": "A String", # The email address of the user or group to which this permission refers.
            "photoLink": "A String", # A link to the user's profile photo, if available.
            "role": "A String", # The role granted by this permission. While new values may be supported in the future, the following are currently allowed:
                # - owner
                # - organizer
                # - fileOrganizer
                # - writer
                # - commenter
                # - reader
            "expirationTime": "A String", # The time at which this permission will expire (RFC 3339 date-time). Expiration times have the following restrictions:
                # - They can only be set on user and group permissions
                # - The time must be in the future
                # - The time cannot be more than a year in the future
            "type": "A String", # The type of the grantee. Valid values are:
                # - user
                # - group
                # - domain
                # - anyone
            "id": "A String", # The ID of this permission. This is a unique identifier for the grantee, and is published in User resources as permissionId.
          },
        ],
        "headRevisionId": "A String", # The ID of the file's head revision. This is currently only available for files with binary content in Drive.
      },
    ],
    "incompleteSearch": True or False, # Whether the search process was incomplete. If true, then some search results may be missing, since all documents were not searched. This may occur when searching multiple Team Drives with the "user,allTeamDrives" corpora, but all corpora could not be searched. When this happens, it is suggested that clients narrow their query by choosing a different corpus such as "user" or "teamDrive".
    "kind": "drive#fileList", # Identifies what kind of resource this is. Value: the fixed string "drive#fileList".
    "nextPageToken": "A String", # The page token for the next page of files. This will be absent if the end of the files list has been reached. If the token is rejected for any reason, it should be discarded, and pagination should be restarted from the first page of results.
  }

Developer Reference: Google Spreadsheet File Object

	These are huge, some key parts have been extracted
	These sheet data include all of the styling and formating 

	Taken from: https://developers.google.com/resources/api-libraries/documentation/sheets/v4/python/latest/sheets_v4.spreadsheets.html#get


	"properties": { # Properties of a sheet. # The properties of the sheet.
          "sheetType": "A String", # The type of sheet. Defaults to GRID.
              # This field cannot be changed once set.
          "index": 42, # The index of the sheet within the spreadsheet.
              # When adding or updating sheet properties, if this field
              # is excluded then the sheet is added or moved to the end
              # of the sheet list. When updating sheet indices or inserting
              # sheets, movement is considered in "before the move" indexes.
              # For example, if there were 3 sheets (S1, S2, S3) in order to
              # move S1 ahead of S2 the index would have to be set to 2. A sheet
              # index update request is ignored if the requested index is
              # identical to the sheets current index or if the requested new
              # index is equal to the current sheet index + 1.
          "title": "A String", # The name of the sheet.
          "gridProperties": { # Properties of a grid. # Additional properties of the sheet if this sheet is a grid.
              # (If the sheet is an object sheet, containing a chart or image, then
              # this field will be absent.)
              # When writing it is an error to set any grid properties on non-grid sheets.
            "columnCount": 42, # The number of columns in the grid.
            "rowGroupControlAfter": True or False, # True if the row grouping control toggle is shown after the group.
            "columnGroupControlAfter": True or False, # True if the column grouping control toggle is shown after the group.
            "frozenRowCount": 42, # The number of rows that are frozen in the grid.
            "frozenColumnCount": 42, # The number of columns that are frozen in the grid.
            "rowCount": 42, # The number of rows in the grid.
            "hideGridlines": True or False, # True if the grid isn't showing gridlines in the UI.
          },


	"data": [ # Data in the grid, if this is a grid sheet.
            # The number of GridData objects returned is dependent on the number of
            # ranges requested on this sheet. For example, if this is representing
            # `Sheet1`, and the spreadsheet was requested with ranges
            # `Sheet1!A1:C10` and `Sheet1!D15:E20`, then the first GridData will have a
            # startRow/startColumn of `0`,
            # while the second one will have `startRow 14` (zero-based row 15),
            # and `startColumn 3` (zero-based column D).
          { # Data in the grid, as well as metadata about the dimensions.
            "startRow": 42, # The first row this GridData refers to, zero-based.
            "startColumn": 42, # The first column this GridData refers to, zero-based.
            "rowData": [ # The data in the grid, one entry per row,
                # starting with the row in startRow.
                # The values in RowData will correspond to columns starting
                # at start_column.
              { # Data about each cell in a row.
                "values": [ # The values in the row, one per column.
                  { # Data about a specific cell.}
                "valueMetadata": [ # Metadata about values in the grouping.
                            { # Metadata about a value in a pivot grouping.
                              "collapsed": True or False, # True if the data corresponding to the value is collapsed.
                              "value": { # The kinds of value that a cell in a spreadsheet can have. # The calculated value the metadata corresponds to.
                                  # (Note that formulaValue is not valid,
                                  #  because the values will be calculated.)
                                "numberValue": 3.14, # Represents a double value.
                                    # Note: Dates, Times and DateTimes are represented as doubles in
                                    # "serial number" format.
                                "boolValue": True or False, # Represents a boolean value.
                                "formulaValue": "A String", # Represents a formula.
                                "stringValue": "A String", # Represents a string value.
                                    # Leading single quotes are not included. For example, if the user typed
                                    # `'123` into the UI, this would be represented as a `stringValue` of
                                    # `"123"`.
                                "errorValue": { # An error in a cell. # Represents an error.
                                    # This field is read-only.
                                  "message": "A String", # A message with more information about the error
                                      # (in the spreadsheet's locale).
                                  "type": "A String", # The type of error.
                                },
                              },
                            },
                          ]



	"effectiveValue": { # The kinds of value that a cell in a spreadsheet can have. # The effective value of the cell. For cells with formulas, this is
                        # the calculated value.  For cells with literals, this is
                        # the same as the user_entered_value.
                        # This field is read-only.
                      "numberValue": 3.14, # Represents a double value.
                          # Note: Dates, Times and DateTimes are represented as doubles in
                          # "serial number" format.
                      "boolValue": True or False, # Represents a boolean value.
                      "formulaValue": "A String", # Represents a formula.
                      "stringValue": "A String", # Represents a string value.
                          # Leading single quotes are not included. For example, if the user typed
                          # `'123` into the UI, this would be represented as a `stringValue` of
                          # `"123"`.
                      "errorValue": { # An error in a cell. # Represents an error.
                          # This field is read-only.
                        "message": "A String", # A message with more information about the error
                            # (in the spreadsheet's locale).
                        "type": "A String", # The type of error.
                      },
                    },
                    "formattedValue": "A String", # The formatted value of the cell.
                        # This is the value as it's shown to the user.
                        # This field is read-only.
                    "userEnteredValue": { # The kinds of value that a cell in a spreadsheet can have. # The value the user entered in the cell. e.g, `1234`, `'Hello'`, or `=NOW()`
                        # Note: Dates, Times and DateTimes are represented as doubles in
                        # serial number format.
                      "numberValue": 3.14, # Represents a double value.
                          # Note: Dates, Times and DateTimes are represented as doubles in
                          # "serial number" format.
                      "boolValue": True or False, # Represents a boolean value.
                      "formulaValue": "A String", # Represents a formula.
                      "stringValue": "A String", # Represents a string value.
                          # Leading single quotes are not included. For example, if the user typed
                          # `'123` into the UI, this would be represented as a `stringValue` of
                          # `"123"`.
                      "errorValue": { # An error in a cell. # Represents an error.
                          # This field is read-only.
                        "message": "A String", # A message with more information about the error
                            # (in the spreadsheet's locale).
                        "type": "A String", # The type of error.
                      }

	"spreadsheetUrl": "A String", # The url of the spreadsheet.
	        # This field is read-only.
	    "spreadsheetId": "A String", # The ID of the spreadsheet.
	        # This field is read-only.

