import os
import sys
import json
import gc
from apiclient.discovery import build
from httplib2 import Http
from oauth2client import file, client, tools

#scopes are space seperated
SCOPES = 'https://www.googleapis.com/auth/drive.readonly https://www.googleapis.com/auth/spreadsheets.readonly'

#Google File types
TYPE_FOLDER = 'application/vnd.google-apps.folder'
TYPE_SPREADSHEET = 'application/vnd.google-apps.spreadsheet'

#Google File object property names
FILE_ID = 'id'
FILE_NAME = 'name'
FILE_MIME_TYPE = 'mimeType'
FILE_FILES = 'files'

#Google Spreadsheet and sheet property names
SPREADSHEET_SHEETS = 'sheets'
SPREADSHEET_URL = 'spreadsheetUrl'
SPREADSHEET_ID = 'spreadsheetId'
SPREADSHEET_PROPERTIES = 'properties'
SPREADSHEET_TITLE = 'title' #contained within properties

SHEET_PROPERTIES = 'properties' #contained within the sheets object

SHEET_TITLE = 'title' #contained within properties
SHEET_ID = 'sheetId' #contained within properties
SHEET_GRID_PROPERTIES = 'gridProperties' #contained within properties

SHEET_COLUMN_COUNT = 'columnCount' #contained within gridProperties
SHEET_ROW_COUNT = 'rowCount' #contained within gridProperties
SHEET_FROZEN_ROW_COUNT = 'frozenRowCount' #contained within gridProperties

SHEET_DATA = 'data' # top level
SHEET_ROW_DATA = 'rowData' #contained within data, an entry per row
SHEET_VALUES = 'values' #contained within rowData, an entry per column
SHEET_EFFECTIVE_VALUE = 'effectiveValue' #contained within the values object, this is the money, the display value in the cell
SHEET_STRING_VALUE = 'stringValue' #contained within effectiveValue, there are a lot of options here to get numbers or booleans, all optional
SHEET_FORMATTED_VALUE = 'formattedValue' #contained within values, if effective isn't correct, I'm going to use this value 



#other Google magic names
DRIVE = 'drive'
DRIVE_VERSION = 'v3'
SHEETS = 'sheets'
SHEETS_VERSION = 'v4'
TOKEN = 'token.json'
CREDENTIALS = 'credentials.json' 


#Our custom JSON names for output files
METADATA = 'metadata'
DATA = 'data'
SCHEMA_KEYS = 'schema'
SCHEMA_COLUMN_INDEX = "schema_column"


#Config JSON Names
CONFIG_GOOGLE_ENTRY_POINT = 'googleParentFolderID'
CONFIG_LOCAL_BASE_PATH = 'localDirectoryBasePath'
CONFIG_RESULT_LIMIT = 'numResults'
CONFIG_GAP_ROWS = 'allowGapRows'
CONFIG_GAP_COLUMNS = 'allowGapColumns'
CONFIG_ENFORCE_ASCII = 'enforceASCII'
CONFIG_JSON_EXT = 'jsonExtension'
CONFIG_DIRECTORY = 'pathSlash'
CONFIG_FILE_ECODING = 'outputEncoding'
CONFIG_LOG_LEVEL = 'logLevel'
CONFIG_LOG_FILE = 'logOutputFile'

#Config values
GOOGLE_ENTRY_POINT = ""
LOCAL_BASE_PATH = "."
RESULT_LIMIT = 1000
GAP_ROWS = True
GAP_COLUMNS = True
ENFORCE_ASCII = False
JSON_EXT = ".json"
DIRECTORY = "/"
FILE_ECODING = 'utf8'
LOG_LEVEL = 0
LOG_FILE = "error.log"

#global directory variable
directoryFileStack = []
#global vars
problemCharacters = set('/\?%:|"><.* ')
errorLogFile = None

#directories
CURRENT_WORKING_DIRECTORY = os.getcwd()
FILE_DIRECTORY = os.path.dirname(os.path.realpath(__file__))

def main():

    #load our configs
    global GOOGLE_ENTRY_POINT
    global LOCAL_BASE_PATH
    global RESULT_LIMIT
    global GAP_ROWS
    global GAP_COLUMNS
    global ENFORCE_ASCII
    global JSON_EXT
    global DIRECTORY
    global FILE_ECODING
    global LOG_LEVEL
    global LOG_FILE
    global TOKEN
    global CREDENTIALS
    
    if len(sys.argv) < 2:
        config = json.loads(open(FILE_DIRECTORY + DIRECTORY + 'config.json').read()) #config should be a dictionary now
    else:
        config = json.loads(open(sys.argv[1]).read()) #So you can pass config file in from any project directory


    if CONFIG_GOOGLE_ENTRY_POINT in config:
        GOOGLE_ENTRY_POINT = config[CONFIG_GOOGLE_ENTRY_POINT]
    else:
        print ("Error, no google file id in config file")

    if CONFIG_LOCAL_BASE_PATH in config:
        LOCAL_BASE_PATH = config[CONFIG_LOCAL_BASE_PATH]
    if CONFIG_RESULT_LIMIT in config:
        RESULT_LIMIT = config[CONFIG_RESULT_LIMIT]
    if CONFIG_GAP_ROWS in config:
        GAP_ROWS = config[CONFIG_GAP_ROWS]
    if CONFIG_GAP_COLUMNS in config:
        GAP_COLUMNS = config[CONFIG_GAP_COLUMNS]
    if CONFIG_ENFORCE_ASCII in config:
        ENFORCE_ASCII = config[CONFIG_ENFORCE_ASCII]
    if CONFIG_JSON_EXT in config:
        JSON_EXT = config[CONFIG_JSON_EXT]
    if CONFIG_DIRECTORY in config:
        DIRECTORY = config[CONFIG_DIRECTORY]
    if CONFIG_FILE_ECODING in config:
        FILE_ECODING = config[CONFIG_FILE_ECODING]
    if CONFIG_LOG_LEVEL in config:
        LOG_LEVEL = config[CONFIG_LOG_LEVEL]
    if CONFIG_LOG_FILE in config:
        LOG_FILE = config[CONFIG_LOG_FILE]
    
    #redefine token + credentials
    TOKEN = FILE_DIRECTORY + "\\" + TOKEN
    CREDENTIALS = FILE_DIRECTORY + "\\" + CREDENTIALS

    #open up a file to log our errors too
    #config file errors will have to be stored
    global errorLogFile
    errorLogFile = open('error.log', 'w+')
    
    if GOOGLE_ENTRY_POINT == "":
        logError(' No google file id in config file')
    
    #Done with config, file cleanup
    config.clear()
    config = None

    # The file token.json stores the user's access and refresh tokens, and is
    # created automatically when the authorization flow completes for the first
    # time.
    store = file.Storage(TOKEN)
    creds = store.get()
    if not creds or creds.invalid:
        flow = client.flow_from_clientsecrets(CREDENTIALS, SCOPES)
        creds = tools.run_flow(flow, store)
    service = build(DRIVE, DRIVE_VERSION, http=creds.authorize(Http()))

    
    # I need multiple lists / recursion, find all folders contained within a folder
    # make queries / do things iterating over all folders discovered
    # build "path strings" which will need folder names, not ids (associative stacks)

    #grab the initial folder, it should always be a folder, maybe I can also support it being a spreadsheet
    initialFileFolder = service.files().get(fileId = GOOGLE_ENTRY_POINT).execute()
    #print(initialFileFolder)

    if initialFileFolder == None:
        logError(' File ' + GOOGLE_ENTRY_POINT + ' Not Found On Drive')
        return

    if initialFileFolder[FILE_MIME_TYPE] == TYPE_SPREADSHEET:
        processSpreadsheet(initialFileFolder);
        return



    directoryFileStack = [initialFileFolder] #other than this first item, the rest will be file JSON object from google
    
    searchDirectory(initialFileFolder)
    
    if errorLogFile != None:
        errorLogFile.close()
    
    print ("Done!")
            
    
#recursion over directories, not individual files, this function should always be passed folder objects

def searchDirectory(directory): #The google drive file object
    
    if directory == None or directory[FILE_MIME_TYPE] != TYPE_FOLDER:
        logError(' Non Folder File Passed to searchDirectory function')
        return

    # We are in a new directory, add it to the stack
    directoryFileStack.append(directory);
    makePath()

    #authenticate, seems like I have to do this in every scope
    store = file.Storage(TOKEN)
    creds = store.get()
    if not creds or creds.invalid:
        flow = client.flow_from_clientsecrets(CREDENTIALS, SCOPES)
        creds = tools.run_flow(flow, store)
    service = build(DRIVE, DRIVE_VERSION, http=creds.authorize(Http()))


    # There could be file or folders in every directory
    
    directoryFILE_ID = "'" + directory[FILE_ID] + "'" #extra quotes are needed for the filter "q" parameter
    folderFiles = service.files().list(spaces = DRIVE, pageSize = RESULT_LIMIT, q = directoryFILE_ID + " in parents and mimeType = '" + TYPE_FOLDER + "'").execute().get(FILE_FILES, [])
    spreadsheetFiles = service.files().list(spaces = DRIVE, pageSize = RESULT_LIMIT, q = directoryFILE_ID + " in parents and mimeType = '" + TYPE_SPREADSHEET + "'").execute().get(FILE_FILES, [])
    tabs = "\t" * len(directoryFileStack)

    logDebug(tabs + "Found " + str(len(folderFiles)) + " folders")


    for f in spreadsheetFiles:
        processSpreadsheet(f)

    if len(folderFiles) != 0:  
        for f in folderFiles:
            logDebug(tabs + f[FILE_NAME] + ',' + f[FILE_MIME_TYPE])
        for f in folderFiles:
            searchDirectory(f)

    #When we are done with the directory, remove it, we should only be visiting every directory once    
    directoryFileStack.pop();


def logError(errorString):
    if errorLogFile != None:
        errorLogFile.write( 'ERROR: ' + errorString + '\r\n')

def logDebug(debugString):    
    if errorLogFile != None:
        errorLogFile.write( 'DEBUG: ' + debugString + '\r\n')

def logWarning(warningString):
    if errorLogFile != None:
        errorLogFile.write( 'WARNING: ' + warningString + '\r\n')

#This is for the local operating system folders, not google drive
#uses the directoryFileStack to make a path
#should also complain about illegal characters
#makes the directory exist on the file system if it does not already
#directories do not need to be made in order, this can make multiple directories
def makePath():

    os.makedirs(getPath(), exist_ok = True)

    return

#This is for the local operating system folders, not google drive
def getPath():
    
    path = LOCAL_BASE_PATH

    for folderFile in directoryFileStack:
        sanatizedName = folderFile[FILE_NAME]
        if set(sanatizedName).isdisjoint(problemCharacters) == False:
            logWarning (" Illegal OS file name characters found in folder name " + folderFile[FILE_NAME])
            #santize the string
            for probelmChar in problemCharacters:
                sanatizedName = sanatizedName.replace(probelmChar, "")

        path = path + DIRECTORY + sanatizedName.strip() #trim whitespace

    return path

def processSpreadsheet(spreadsheetFile): #The google drive file object, not the file data

    #this actually doesn't do too much, it just loops over the sheets contained within the spreadsheet
    #we will make another folder for with the spreadsheet title

    if spreadsheetFile == None or spreadsheetFile[FILE_MIME_TYPE] != TYPE_SPREADSHEET:
        return

    print ('%s, %s' % (spreadsheetFile[FILE_NAME], spreadsheetFile[FILE_MIME_TYPE]))
    directoryFileStack.append(spreadsheetFile);
    makePath()


    #actually need to retrieve the full file, not just the drive file properties, which is what spreadsheetFile is

    store = file.Storage(TOKEN)
    creds = store.get()
    if not creds or creds.invalid:
        flow = client.flow_from_clientsecrets(CREDENTIALS, SCOPES)
        creds = tools.run_flow(flow, store)
    service = build(SHEETS, SHEETS_VERSION, http=creds.authorize(Http()))

    # Call the Sheets API
    sheet = service.spreadsheets()
    #print (spreadsheetFile[FILE_ID])
    result = sheet.get(spreadsheetId = spreadsheetFile[FILE_ID], includeGridData = True).execute()

    while len(result[SPREADSHEET_SHEETS]) > 0:
        processSheet(result[SPREADSHEET_SHEETS].pop(), result[SPREADSHEET_URL])
    
    directoryFileStack.pop();
    #free memory of result
    del result
    #garbage collect
    gc.collect()


    

#given a list of keys
#an associative list of indicies for those keys
#and an index
#return the key name if the index matches in associativeIndicies
def indexToKeyValue(keys, associativeIndicies, index):

    i = 0
    for indexValue in associativeIndicies:
        if indexValue == index:
            return keys[i]
        i += 1

    return None


def processSheet(sheetDataObject, url): #the JSON representation of a tab within a spreadsheet

    #make a json file at the current path using the name of this tab / sheet

    #Even simple sheets are just a wall of text
    #print (sheetDataObject)

    #print ("Imma sheet within a spreadheet %s, %s" % (sheetDataObject[SHEET_PROPERTIES][SHEET_TITLE], url))
    
    #Basic high level structure, our json files are going to have two top level objects
    #metadata object - for general file data, particularly URLs for linking, maybe other file statstics
    #data object - actual content contained within cells

    #We just make a python dictionary object, json.dump will take care of the rest for us

    #print (sheetDataObject[SHEET_DATA])

    filedata = {}
    filedata[METADATA] = {}
    filedata[METADATA][SPREADSHEET_URL] = url
    filedata[METADATA][SHEET_TITLE] = sheetDataObject[SHEET_PROPERTIES][SHEET_TITLE]
    filedata[METADATA][SHEET_ID] = sheetDataObject[SHEET_PROPERTIES][SHEET_ID]

    #these row and column counts include empty data, doubtful that they will really be useful in any meaningful way
    filedata[METADATA][SHEET_COLUMN_COUNT] = sheetDataObject[SHEET_PROPERTIES][SHEET_GRID_PROPERTIES][SHEET_COLUMN_COUNT]
    filedata[METADATA][SHEET_ROW_COUNT] = sheetDataObject[SHEET_PROPERTIES][SHEET_GRID_PROPERTIES][SHEET_ROW_COUNT]
        
    #frozen row count does not always exist if it is not used
    if SHEET_FROZEN_ROW_COUNT in sheetDataObject[SHEET_PROPERTIES][SHEET_GRID_PROPERTIES]:
        filedata[METADATA][SHEET_FROZEN_ROW_COUNT] = sheetDataObject[SHEET_PROPERTIES][SHEET_GRID_PROPERTIES][SHEET_FROZEN_ROW_COUNT]
    else:
        filedata[METADATA][SHEET_FROZEN_ROW_COUNT] = 0
    
    

    filedata[DATA] = []


    #first row should be locked, it is our database schema, which will be our dictionary keys
    #we need to scan the first row to figure out our keys
    
    firstRow = []
    firstRowColumnIndex = [] #because gap columns are allowed, and google data is sparse
    rowCounter = 0;
    columnCounter = 0;

    #make sure there are actually gridData object and rows in that gridData
    if (SHEET_DATA in sheetDataObject) and len(sheetDataObject[SHEET_DATA]) > 0:

        #a completely empty sheet will have no rows
        if SHEET_ROW_DATA in sheetDataObject[SHEET_DATA][0]:
            for rowItem in sheetDataObject[SHEET_DATA][0][SHEET_ROW_DATA]: #we are always getting the entire spreadsheet as one GridData object, hence the [0] index
                
                if SHEET_VALUES in rowItem == False:
                    if GAP_ROWS == False:
                        break
                    else:
                        rowCounter+=1;
                        continue;

                if rowCounter != 0 and rowCounter < filedata[METADATA][SHEET_FROZEN_ROW_COUNT]:
                    rowCounter+=1;
                    continue;


                #a completely empty row will have no columns, this might be a good time to break, or have a config to allow gap rows
                #dictionary object for this row
                rowDictionary = {}
                columnCounter = 0;

                if SHEET_VALUES in rowItem:

                    for columnItem in rowItem[SHEET_VALUES]:
                        #print (rowCounter, columnCounter, columnItem)
                        
                        if (SHEET_FORMATTED_VALUE in columnItem) == False or columnItem[SHEET_FORMATTED_VALUE] == None or columnItem[SHEET_FORMATTED_VALUE] == "":
                            columnCounter+=1;
                            if rowCounter == 0 and GAP_COLUMNS == False:
                                break;
                            else:
                                continue;

                        if rowCounter == 0:
                            #add value to our key list
                            firstRow.append(columnItem[SHEET_FORMATTED_VALUE])
                            firstRowColumnIndex.append(columnCounter)

                        else:
                            #add to file data
                            #if there is a key for this column, need to resolve if there is a key value for this index
                            #find method errors if not found -_-
                            keyValue = indexToKeyValue(firstRow, firstRowColumnIndex, columnCounter)
                            if keyValue != None:
                                rowDictionary[keyValue] = columnItem[SHEET_FORMATTED_VALUE]

                        columnCounter+=1;
                elif GAP_ROWS == False:
                    #row was empty, empty rows ends the data
                    break;

                if len(rowDictionary) != 0:
                    filedata[DATA].append(rowDictionary)
                rowCounter+=1;


    filedata[METADATA][SCHEMA_KEYS] = firstRow
    filedata[METADATA][SCHEMA_COLUMN_INDEX] = firstRowColumnIndex
    
    #print (firstRow)
    
    #skip the rest of the frozen rows, start making data object

    #I still have to worry about dealing with existing files
    newFilePath = getPath() + DIRECTORY +  sheetDataObject[SHEET_PROPERTIES][SHEET_TITLE] + JSON_EXT
    if os.path.exists(newFilePath):
        os.remove(newFilePath)


    with open(newFilePath, 'w', encoding = 'utf8') as outfile:
        json.dump(filedata, outfile, indent = 4, ensure_ascii = ENFORCE_ASCII)

    #aggressively free memory
    filedata.clear()



if __name__ == '__main__':
    main()